#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from dhub.dataset import Dataset
from dhub.dhubrc import dhubrc
from dhub.wrapper.api_wrapper import APIWrapper

__author__ = 'Iván de Paz Centeno'

class Datasets(APIWrapper):
    def __init__(self, token_id=None):

        try:
            token = dhubrc.lookup_token(token_id)
        except KeyError as ex:
            token = token_id

        super().__init__(token)
        self.refresh()

    def __len__(self):
        return len(self.datasets)

    def find_closest(self, item):
        for d in self.keys():
            if item in d:
                return d
        return None

    def __getitem__(self, item) -> Dataset:
        result = None
        try:
            index = int(item)
            result = self.values()[index]
        except ValueError as ex:
            pass

        if result is None:
            # Let's do smart things over here.
            if item not in self.keys():
                closest_name = self.find_closest(item)
                if closest_name is not None:
                    item = closest_name

            result = self.datasets[item]

            if type(result) is dict:
                result = Dataset.from_dict(**result)

        return result

    def add_dataset(self, url_prefix, title, description, reference, tags) -> Dataset:
        result = self._post_json("datasets", json_data={
            'title': title,
            'description': description,
            'tags': tags,
            'url_prefix': url_prefix,
            'reference': reference
        })

        self.refresh()
        return self[result['url_prefix']]

    def __setitem__(self, key, dataset):
        if dataset.api_url in self.keys():
            return

        self._post_json("tokens/{}/link/{}".format(self.token, dataset.get_url_prefix()))
        self.refresh()

    def __delitem__(self, key):
        try:
            index = int(key)
            item_key = self.keys()[index]
        except ValueError as ex:
            item_key = None
            pass

        if item_key is None:
            if key not in self.keys():
                return
            else:
                item_key = key

        self._delete_json("datasets/{}".format(item_key))
        self.refresh()

    def __iter__(self):
        """
        :rtype: Dataset
        :return:
        """
        for key in self.keys():
            yield self[key]

    def keys(self):
        return list(self.datasets.keys())

    def values(self):
        return [s for s in self]

    def refresh(self):
        self.datasets = {d['url_prefix']: dict(definition=d, token=self.token, token_info=self.token_info, server_info=self.server_info) for d in self._get_json("datasets")}

    def __str__(self):
        return str(self.keys())

    def __repr__(self):
        return "[{}] {}".format(self.api_url, str(self))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from concurrent.futures import ThreadPoolExecutor
from dhub.config import now, CACHE_TIME

from dhub.wrapper.api_wrapper import APIWrapper
from dhub.wrapper.smart_updater import AsyncSmartUpdater

__author__ = 'Iván de Paz Centeno'


upload_pool = ThreadPoolExecutor(4)


class Element(APIWrapper):

    def __init__(self, title, description, tags, http_ref, id=None, dataset_owner=None, token=None,
                 binary_interpreter=None, token_info=None, server_info=None, smart_updater:AsyncSmartUpdater=None):
        super().__init__(token, token_info=token_info, server_info=server_info)
        self.data = {'title': title, 'description': description, 'tags': tags, 'http_ref': http_ref}
        self.has_content = False
        self.dataset_owner = dataset_owner
        self.binary_interpreter = binary_interpreter
        self.token = token
        self._id = id
        self.cached_content = None
        self.cached_content_time = now()
        self.content_promise = None
        self.smart_updater = smart_updater

    def _retrieve_content(self):

        if self.smart_updater is not None and self.smart_updater.is_content_update_queued([self.get_id()]):
            self.smart_updater.wait_for_elements_content_update([self.get_id()])

        content = self._get_binary("datasets/{}/elements/{}/content".format(self.dataset_owner.get_url_prefix(), self._id))
        return {self._id: content}

    def _upload_content(self, content):
        return self._put_binary("datasets/{}/elements/{}/content".format(self.dataset_owner.get_url_prefix(), self._id), binary=content)

    def get_title(self):
        return self.data['title']

    def get_description(self):
        return self.data['description']

    def get_tags(self):
        return self.data['tags']

    def get_ref(self):
        return self.data['http_ref']

    def get_id(self):
        return self._id

    def set_title(self, new_title):
        self.data['title'] = new_title
        self.update()

    def set_description(self, new_description):
        self.data['description'] = new_description
        self.update()

    def set_tags(self, new_tags):
        self.data['tags'] = new_tags
        self.update()

    def set_ref(self, new_http_ref):
        self.data['http_ref'] = new_http_ref
        self.update()

    def get_content(self, interpret=True):
        if not self.has_content:
            return None

        if self.content_promise is not None:
            self.cached_content = self.content_promise.result()[self._id]
            self.cached_content_time = now()
            self.content_promise = None

        if (now() - self.cached_content_time).total_seconds() > CACHE_TIME:
            self.cached_content = None

        if self.cached_content is None:
            content = self._retrieve_content()[self._id]
        else:
            content = self.cached_content

        if self.binary_interpreter is not None and interpret:
            content = self.binary_interpreter.cipher(content)

        return content

    def set_content(self, content, interpret=True):
        if self.binary_interpreter is not None and interpret:
            content = self.binary_interpreter.decipher(content)
        else:
            if type(content) is not bytes:
                raise Exception("Bytes are required as content.")

        if content == self.cached_content:
            return False

        if self.smart_updater is not None:
            self.smart_updater.queue_content_update("datasets/{}/elements/content".format(self.dataset_owner.get_url_prefix()), self.get_id(), content)
        else:
            upload_pool.submit(self._upload_content, content)

        self.cached_content = content
        self.has_content = True
        self.cached_content_time = now()

        if self.content_promise:
            self.content_promise = None

        return True

    def update(self):
        if self.smart_updater is not None:
            self.smart_updater.queue_update("datasets/{}/elements/bundle".format(self.dataset_owner.get_url_prefix()), self.get_id(), self.data)

        else:
            self._patch_json("datasets/{}/elements/{}".format(self.dataset_owner.get_url_prefix(), self._id),
                             json_data=self.data)
            self.refresh()

    def __str__(self):
        return "{} {}".format(self._id, str(self.data))

    def __repr__(self):
        return "[{}] {}".format(self._id, str(self.data))

    @classmethod
    def from_dict(cls, definition, dataset_owner, token, binary_interpreter=None, token_info=None, server_info=None,
                  smart_updater=None):

        element = cls(definition['title'], definition['description'], definition['tags'],
                      definition['http_ref'], id=definition['_id'], token=token,
                      binary_interpreter=binary_interpreter, dataset_owner=dataset_owner, token_info=token_info, server_info=server_info,
                      smart_updater=smart_updater)

        element.comments_count = definition['comments_count']
        element.has_content = definition['has_content']
        return element


    def refresh(self):
        definition = self._get_json("datasets/{}/elements/{}".format(self.dataset_owner.get_url_prefix(), self.get_id()))

        self.data = {k: definition[k] for k in ['title', 'description', 'tags', 'http_ref']}
        self.comments_count = definition['comments_count']
        self.has_content = definition['has_content']
        self.cached_content = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import concurrent
from concurrent.futures import ThreadPoolExecutor
import os
from queue import Queue, Empty
import threading
from time import sleep
from pyzip import PyZip

__author__ = 'Iván de Paz Centeno'

pool = ThreadPoolExecutor(4)
UPDATE_INTERVAL = 1  # seconds


class AsyncSmartUpdater(object):

    __tasks_pending = 0

    def __init__(self, server_info, api_wrapper_owner):
        self.server_info = server_info
        self.api_wrapper_owner = api_wrapper_owner
        self.content_put_queue = Queue()
        self.element_update_queue = Queue()
        self.queues_priorities = [self.element_update_queue, self.content_put_queue]
        self.queues_cache = {'element_update': {}, 'content_update': {}}
        self.__exit = False
        self.__cancel_pending_jobs = False
        self.lock = threading.Lock()
        self.thread = threading.Thread(target=self._thread_func, daemon=True)
        self.thread.start()

    @property
    def _exit(self):
        with self.lock:
            exit_value = self.__exit
        return exit_value

    @property
    def tasks_pending(self):
        with self.lock:
            return self.__tasks_pending

    @tasks_pending.setter
    def tasks_pending(self, __tasks_pending):
        with self.lock:
            self.__tasks_pending = __tasks_pending

    def _increase_tasks_pending_counter(self, by=1):
        with self.lock:
            self.__tasks_pending += by

    def _decrease_tasks_pending_counter(self, by=1):
        with self.lock:
            self.__tasks_pending -= by

    @property
    def _cancel_pending_jobs(self):
        with self.lock:
            cancel_pending_jobs = self.__cancel_pending_jobs
        return cancel_pending_jobs

    @_cancel_pending_jobs.setter
    def _cancel_pending_jobs(self, do__cancel_pending_jobs=True):
        with self.lock:
            self.__cancel_pending_jobs = do__cancel_pending_jobs

    def __queue_with_higher_priority_waiting(self, queue):
        """
        Checks if is there any queue with higher priority waiting than the specified queue.
        :param queue: queue to check from
        :return: boolean True if there is at least one queue with higher priority waiting. False otherwise.
        """
        index = 0
        current_queue_index = self.queues_priorities.index(queue)

        while index < current_queue_index and self.queues_priorities[index].qsize() == 0:
            index += 1

        return index != len(self.queues_priorities) and index < current_queue_index

    def __any_queue_with_elements(self):
        return any([queue.qsize() > 0 for queue in self.queues_priorities])

    def is_element_update_queued(self, elements_ids):
        return any(element in self.queues_cache['element_update'] for element in elements_ids)

    def is_content_update_queued(self, elements_ids):
        return any(element in self.queues_cache['content_update'] for element in elements_ids)

    def wait_for_elements_update(self, elements_ids):
        events_list = []

        with self.lock:
            for element_id in elements_ids:
                if element_id not in self.queues_cache['element_update']:
                    continue
                if self.queues_cache['element_update'][element_id] is None:
                    event = threading.Event()
                    self.queues_cache['element_update'][element_id] = event
                else:
                    event = self.queues_cache['element_update'][element_id]

                events_list.append(event)

        launch_exception = None

        try:
            # Now we wait for the events
            for event in events_list:
                event = event
                """
                :type : threading.Event
                """
                if not event.wait(100):
                    raise Exception("Time out while waiting for the elements.")
        except Exception as ex:
            launch_exception = ex
        finally:
            # do some cleanup
            with self.lock:
                for element_id in elements_ids:
                    if element_id in self.queues_cache['element_update']:
                        del self.queues_cache['element_update'][element_id]

        if launch_exception is not None:
            raise launch_exception

    def wait_for_elements_content_update(self, elements_ids):
        events_list = []

        with self.lock:
            for element_id in elements_ids:
                if element_id not in self.queues_cache['content_update']:
                    continue
                if self.queues_cache['content_update'][element_id] is None:
                    event = threading.Event()
                    self.queues_cache['content_update'][element_id] = event
                else:
                    event = self.queues_cache['content_update'][element_id]

                events_list.append(event)

        launch_exception = None

        try:
            # Now we wait for the events
            for event in events_list:
                event = event
                """
                :type : threading.Event
                """
                if not event.wait(100):
                    raise Exception("Time out while waiting for the elements.")
        except Exception as ex:
            launch_exception = ex
        finally:
            # do some cleanup
            with self.lock:
                for element_id in elements_ids:
                    if element_id in self.queues_cache['content_update']:
                        del self.queues_cache['content_update'][element_id]

        if launch_exception is not None:
            raise launch_exception

    def _thread_func(self):
        ps = self.server_info['Page-Size']
        queue_types = {self.content_put_queue: 'binary', self.element_update_queue: 'json'}

        while not self._exit or (not self._cancel_pending_jobs and self.__any_queue_with_elements()):
            # Let's gather all the updates from the queue to serialize them
            sleep(UPDATE_INTERVAL)
            for queue in self.queues_priorities:

                if self.__queue_with_higher_priority_waiting(queue):
                    break  # This forces the loop to start running through the most priority queues again.

                gathered_elements = []
                futures = {}
                queue_empty = False

                while not queue_empty:

                    # Check that queues with highest priority are not waiting
                    if self.__queue_with_higher_priority_waiting(queue):
                        break  # This forces the loop to start running through the most priority queues again.

                    try:
                        element = queue.get(block=False)
                        gathered_elements.append(element)
                    except Empty as ex:
                        queue_empty = True

                    # This is the smart action: we combine several requests into one
                    if len(gathered_elements) > ps-1:
                        futures[pool.submit(self.__do_update, queue_types[queue], gathered_elements)] = len(gathered_elements)
                        gathered_elements = []

                if len(gathered_elements) > 0:
                    futures[pool.submit(self.__do_update, queue_types[queue], gathered_elements)] = len(gathered_elements)

                for future in concurrent.futures.as_completed(futures):
                    self._decrease_tasks_pending_counter(futures[future])

    def queues_busy(self):
        return self.tasks_pending > 0

    def __do_update(self, request_kind, elements):
        if len(elements) > 0:
            url = elements[0][0]
        else:
            url = None

        kwargs_list = {}
        for element in elements:
            kwargs_list[element[1]] = element[2]

        if len(kwargs_list) == 0:
            return None

        if request_kind == "json":
            self.api_wrapper_owner._patch_json(url, extra_data=None, json_data={'elements': kwargs_list})
            for [_, element_id, _] in elements:
                event = self.queues_cache['element_update'][element_id]
                """
                :type : threading.Event
                """

                if event is not None:
                    event.set()
                else:
                    with self.lock:
                        del self.queues_cache['element_update'][element_id]

        else: # request_kind == "binary":
            content = PyZip(kwargs_list).to_bytes()
            self.api_wrapper_owner._put_binary(url, extra_data=None, binary=content)
            for [_, element_id, _] in elements:
                event = self.queues_cache['content_update'][element_id]
                """
                :type : threading.Event
                """

                if event is not None:
                    event.set()
                else:
                    with self.lock:
                        del self.queues_cache['content_update'][element_id]

        return True

    def queue_update(self, url, element_id, kwargs):
        self._increase_tasks_pending_counter()
        with self.lock:
            self.queues_cache["element_update"][element_id] = None
        self.element_update_queue.put([url, element_id, kwargs])

    def queue_content_update(self, url, element_id, content):
        self._increase_tasks_pending_counter()
        with self.lock:
            self.queues_cache["content_update"][element_id] = None
        self.content_put_queue.put([url, element_id, content])

    def stop(self, cancel_pending_jobs=False):
        self.__exit = True
        self.__cancel_pending_jobs = cancel_pending_jobs
        self.thread.join(10)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.


from concurrent.futures import ThreadPoolExecutor
import csv
import json
import os
from time import sleep
from pyfolder import PyFolder
from pyzip import PyZip
from dhub.config import now, CACHE_TIME, segments
from dhub.element import Element
from dhub.wrapper.api_wrapper import APIWrapper
from dhub.interpreters.interpreter import Interpreter
from dhub.wrapper.smart_updater import AsyncSmartUpdater

__author__ = 'Iván de Paz Centeno'

pool_keys = ThreadPoolExecutor(4)
pool_content = ThreadPoolExecutor(4)


class Dataset(APIWrapper):
    def __init__(self, url_prefix: str, title: str, description: str, reference: str, tags: list, token: str=None,
                 binary_interpreter: Interpreter=None, token_info: dict=None, server_info: dict=None,
                 use_smart_updater: AsyncSmartUpdater=True):
        self.data = {}

        self.binary_interpreter = binary_interpreter
        """:type : Interpreter"""

        if "/" in url_prefix:
            self.data['url_prefix'] = url_prefix
        else:
            if token is not None:
                token_prefix = token.get_prefix()
                self.data['url_prefix'] = '{}/{}'.format(token_prefix, url_prefix)

        self.data['title'] = title
        self.data['description'] = description
        self.data['tags'] = tags
        self.data['reference'] = reference
        self.data['size'] = 0
        self.elements_count = 0
        self.comments_count = 0
        self.page_cache = {}
        self.last_cache_time = now()
        super().__init__(token, token_info=token_info, server_info=server_info)

        # Server_info is only available after super() init.
        if use_smart_updater:
            self.smart_updater = AsyncSmartUpdater(self.server_info, self)
            """
            :type : AsyncSmartUpdater
            """
        else:
            self.smart_updater = None
            """
            :type : AsyncSmartUpdater
            """

    def __repr__(self):

        return "Dataset {} ({} elements); tags: {}; description: {}; reference: {}; fork_father: {}, fork_count: {}, size: {}".format(self.get_title(), len(self),
                                                                                           self.get_tags(),
                                                                                           self.get_description(),
                                                                                           self.get_reference(),
                                                                                           self.get_fork_father(),
                                                                                           self.get_fork_count(),
                                                                                           self.get_size())


    def set_binary_interpreter(self, binary_interpreter):
        self.binary_interpreter = binary_interpreter

    def get_fork_father(self):
        return self.data['fork_father']

    def get_fork_count(self):
        return self.data['fork_count']

    def get_url_prefix(self):
        return self.data['url_prefix']

    def get_description(self):
        return self.data['description']

    def get_title(self):
        return self.data['title']

    def get_tags(self):
        return self.data['tags']

    def get_size(self):
        return self.data['size']

    def get_reference(self):
        return self.data['reference']

    def set_description(self, new_desc):
        self.data['description'] = new_desc

    def set_title(self, new_title):
        self.data['title'] = new_title

    def set_tags(self, new_tags):
        self.data['tags'] = new_tags

    def set_reference(self, new_reference):
        self.data['reference'] = new_reference

    def update(self):
        self._patch_json("/datasets/{}".format(self.get_url_prefix()),
                         json_data={k: v for k, v in self.data.items() if k != "url_prefix"})

    @classmethod
    def from_dict(cls, definition, token, binary_interpreter=None, token_info=None, server_info=None):

        dataset = cls(definition['url_prefix'], definition['title'], definition['description'], definition['reference'],
                      definition['tags'], token=token, binary_interpreter=binary_interpreter, token_info=token_info,
                      server_info=server_info)
        dataset.data['fork_count'] = definition['fork_count']
        dataset.data['fork_father'] = definition['fork_father']
        dataset.data['size'] = definition['size']
        dataset.comments_count = definition['comments_count']
        dataset.elements_count = definition['elements_count']
        return dataset

    def add_element(self, title: str, description: str, tags: list, http_ref: str, content, interpret=True) -> Element:

        if type(content) is str:
            # content is a URI
            if not os.path.exists(content):
                raise Exception("content must be a binary data or a URI to a file.")

            with open(content, "rb") as f:
                content_bytes = f.read()

            if self.binary_interpreter is None:
                content = content_bytes
            else:
                content = self.binary_interpreter.cipher(content_bytes)

        result = self._post_json("datasets/{}/elements".format(self.get_url_prefix()), json_data={
            'title': title,
            'description': description,
            'tags': tags,
            'http_ref': http_ref,
        })

        self.refresh()
        element = self[result]
        element.set_content(content, interpret)
        return element

    def add_elements(self, add_element_kwargs_list: list) -> list:
        # we need to preprocess the argument
        post_kwargs = []
        content_list = []
        for element_kwargs in add_element_kwargs_list:
            content = element_kwargs['content']

            if type(content) is str:
                # content is a URI
                if not os.path.exists(content):
                    raise Exception("content must be a binary data or a URI to a file.")

                with open(content, "rb") as f:
                    content_bytes = f.read()

                if self.binary_interpreter is None:
                    content = content_bytes
                else:
                    content = self.binary_interpreter.cipher(content_bytes)

            content_list.append(content)

            post_kwargs.append({'title': element_kwargs['title'],
                                'description': element_kwargs['description'],
                                'tags': element_kwargs['tags'],
                                'http_ref': element_kwargs['http_ref']})

        result = self._post_json("datasets/{}/elements/bundle".format(self.get_url_prefix()), json_data={
            'elements': post_kwargs
        })

        self.refresh()
        elements = [Element.from_dict(element, self, self.token, self.binary_interpreter, token_info=self.token_info,
                                  server_info=self.server_info, smart_updater=self.smart_updater) for element in result]

        for element, content in zip(elements, content_list):
            element.set_content(content)

        return elements

    def _request_segment(self, ids):
        results = self._get_json("datasets/{}/elements/bundle".format(self.get_url_prefix()),
                                 json_data={'elements': ids})

        # Warning: what if 'result' does not have the elements ordered in the same way as 'ids'?
        # Todo: reorder 'elements' to match the order of 'ids'
        elements = [
            Element.from_dict(result, self, self.token, self.binary_interpreter, token_info=self.token_info,
                              server_info=self.server_info, smart_updater=self.smart_updater)
            for result in results
            ]

        future = pool_content.submit(self.__retrieve_segment_contents, ids)

        for element in elements:
            element.content_promise = future

        return elements

    def __retrieve_segment_contents(self, ids):

        if self.smart_updater is not None and self.smart_updater.is_content_update_queued(ids):
            self.smart_updater.wait_for_elements_content_update(ids)

        packet_bytes = self._get_binary("datasets/{}/elements/content".format(self.get_url_prefix()),
                                        json_data={'elements': ids})
        packet = PyZip().from_bytes(packet_bytes)

        return dict(packet)

    def __wait_for_elements_ready(self, ids):
        if self.smart_updater is not None:
            if self.smart_updater.is_element_update_queued(ids):
                self.smart_updater.wait_for_elements_update(ids)

    def __getitem__(self, key):
        if type(key) is not slice and len(str(key)) < 8:
            key = int(key)
            key = slice(key, key + 1, 1)

        elements = []

        if type(key) is slice:
            step = key.step
            start = key.start
            stop = key.stop
            if key.step is None: step = 1
            if key.stop is None: stop = len(self)
            if key.stop < 0: stop = len(self) - stop

            ps = self.server_info['Page-Size']

            ids = [self._get_key(i) for i in range(start, stop, step)]

            self.__wait_for_elements_ready(ids)

            futures = [pool_keys.submit(self._request_segment, ids) for segment in segments(ids, ps)]

            elements = []

            for future in futures:
                elements += future.result()

        elif type(key) is str:
            try:
                self.__wait_for_elements_ready([key])
                response = self._get_json("datasets/{}/elements/{}".format(self.get_url_prefix(), key))
                element = Element.from_dict(response, self, self.token, self.binary_interpreter,
                                            token_info=self.token_info, server_info=self.server_info,
                                            smart_updater=self.smart_updater)
                element.content_promise = pool_content.submit(element._retrieve_content)
                elements = [element]

            except Exception as ex:
                elements = []

        else:
            raise KeyError("Type of key not allowed.")

        if len(elements) > 1:
            result = elements
        elif len(elements) == 1:
            result = elements[0]
        else:
            raise KeyError("{} not found".format(key))

        return result

    def __delitem__(self, key):
        if type(key) is not slice and len(str(key)) < 8:
            key = int(key)
            key = slice(key, key + 1, 1)

        ids = []

        if type(key) is slice:
            step = key.step
            start = key.start
            stop = key.stop
            if key.step is None: step = 1
            if key.stop is None: stop = len(self)
            if key.stop < 0: stop = len(self) - stop

            ps = self.server_info['Page-Size']

            ids = [self._get_key(i) for i in range(start, stop, step)]

        elif type(key) is str:
                ids = [key]

        if len(ids) > 1:
            result = self._delete_json("datasets/{}/elements/bundle".format(self.get_url_prefix()), json_data={'elements': ids})
        elif len(ids) == 1:
            result = self._delete_json("datasets/{}/elements/{}".format(self.get_url_prefix(), ids[0]))
        else:
            raise KeyError("{} not found".format(key))

        self.refresh()
        return result

    def _get_elements(self, page, filter_options=None):
        return [Element.from_dict(element, self, self.token, self.binary_interpreter, token_info=self.token_info,
                                  server_info=self.server_info, smart_updater=self.smart_updater) for element in
                self._get_json("datasets/{}/elements".format(self.get_url_prefix()), extra_data={'page': page}, json_data={'options': filter_options})]

    def fork(self, new_prefix:str, title:str, description:str, tags:list, reference:str, destination, options:dict=None):
        arguments = {
                        'title': title,
                        'description': description,
                        'tags': tags,
                        'url_prefix': new_prefix,
                        'reference': reference,
                        'options': options
                    }
        result = self._post_json("datasets/{}/fork/{}".format(self.get_url_prefix(), destination.token),
                        json_data=arguments)

        destination.refresh()
        return destination[result['url_prefix']]

    def filter_iter(self, options=None, cache_content=False):
        """
        :return:
        """
        if options is None:
            options = {}

        ps = self.server_info['Page-Size']
        number_of_pages = len(self) // ps + int(len(self) % ps > 0)

        buffer = None

        for page in range(number_of_pages):

            if buffer is None:
                buffer = pool_keys.submit(self._get_elements, page, options)

            buffer2 = pool_keys.submit(self._get_elements, page + 1, options)

            elements = buffer.result()

            if len(elements) == 0:
                break

            if cache_content:
                future = pool_content.submit(self.__retrieve_segment_contents, [element.get_id() for element in elements])
            else:
                future = None

            for element in elements:
                if cache_content:
                    element.content_promise = future
                yield element

            buffer = buffer2

    def __iter__(self) -> Element:
        for element in self.filter_iter():
            yield element

    def _get_key(self, key_index):
        ps = int(self.server_info['Page-Size'])
        key_page = key_index // ps
        index = key_index % ps

        try:
            if (now() - self.last_cache_time).total_seconds() > CACHE_TIME:
                self.page_cache.clear()

            page = self.page_cache[key_page]

        except KeyError as ex:
            # Cache miss
            page = self._get_json("datasets/{}/elements".format(self.get_url_prefix()), extra_data={'page': key_page})
            self.page_cache[key_page] = page
            self.last_cache_time = now()

        return page[index]['_id']

    def keys(self, page=-1):
        if page == -1:
            data = [self._get_key(i) for i in range(len(self))]
        else:
            data = [element['_id'] for element in
                    self._get_json("datasets/{}/elements".format(self.get_url_prefix()), extra_data={'page': page})]

        return data

    def __len__(self):
        return self.elements_count

    def __str__(self):
        data = dict(self.data)
        data['num_elements'] = len(self)
        url_prefix = data['url_prefix']
        del data['url_prefix']
        order_keys = ['title', 'num_elements', 'reference', 'tags', 'fork_father', 'fork_count', 'description']

        result = "[{}]".format(url_prefix)
        result += " {}"
        content ="{"
        first = True
        for key in order_keys:
            if first:
                first = False
            else:
                content += ", "

            if key == 'tags':
                content += "'{}': {}".format(key, data[key])
            else:
                content += "'{}': \'{}\'".format(key, data[key])

        content += ", '{}': {} MB".format('size', round(int(data['size'])/1024/1024, 3))
        content += "}"

        return result.format(content)

    def save_to_folder(self, folder, metadata_format="json", elements_extension=None, use_numbered_ids=False,
                       only_metadata=False):
        try:
            os.mkdir(folder)
        except Exception as ex:
            pass

        format_saver = {
            "csv": self.__save_csv,
            "json": self.__save_json
        }

        if elements_extension is not None and elements_extension.startswith("."):
            elements_extension = elements_extension[1:]

        if metadata_format not in format_saver:
            raise Exception("format {} for metadata not supported.".format(metadata_format))

        pyfolder = PyFolder(folder)

        print("Collecting elements...")
        metadata = {}
        id = -1
        count = len(self)
        it = -1

        for element in self.filter_iter(cache_content=not only_metadata):
            it += 1

            print("\rProgress: {}%".format(round(it / (count + 0.0001) * 100, 2)), end="", flush=True)

            if use_numbered_ids:
                id += 1
            else:
                id = element.get_id()

            if elements_extension is None:
                element_id = id
            else:
                element_id = "{}.{}".format(id, elements_extension)

            metadata[element_id] = {
                'id': element.get_id(),
                'title': element.get_title(),
                'description': element.get_description(),
                'http_ref': element.get_ref(),
                'tags': element.get_tags(),
            }

            if not only_metadata:
                pyfolder[os.path.join("content", element_id)] = element.get_content(interpret=False)

        print("\rProgress: 100%", end="", flush=True)

        format_saver[metadata_format](pyfolder, metadata)
        print("\nSaved metadata in format {}".format(metadata_format))
        print("Finished")

    def __save_json(self, pyfolder:PyFolder, metadata):
        pyfolder["metadata.json"] = metadata
        data = dict(self.data)
        data['num_elements'] = len(self)
        data['tags'] = data['tags']
        pyfolder["dataset_info.json"] = data

    def __save_csv(self, pyfolder:PyFolder, metadata):
        with open(os.path.join(pyfolder.folder_root, "metadata.csv"), 'w', newline="") as f:
            writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            headers = ['file_name', 'id', 'title', 'description', 'http_ref', 'tags']
            writer.writerow(headers)

            for k, v in metadata.items():
                v['tags'] = "'{}'".format(";".join(v['tags']))
                writer.writerow([k] + [v[h] for h in headers if h != 'file_name'])

        with open(os.path.join(pyfolder.folder_root, "dataset_info.csv"), "w", newline="") as f:
            writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            data = dict(self.data)
            data['num_elements'] = len(self)
            data['tags'] = "'{}'".format(";".join(data['tags']))
            headers = ['url_prefix', 'title', 'description', 'num_elements', 'reference', 'tags']
            writer.writerow(headers)
            writer.writerow([data[h] for h in headers])

    def refresh(self):
        dataset_data = self._get_json("datasets/{}".format(self.get_url_prefix()))
        self.elements_count = dataset_data['elements_count']
        self.comments_count = dataset_data['comments_count']
        self.data = {k: dataset_data[k] for k in ['url_prefix', 'title', 'description', 'reference', 'tags',
                                                  'fork_count', 'fork_father', 'size']}

    def clear(self):
        ps = self.server_info['Page-Size']

        for segment in segments(self.keys(), ps):
            self._delete_json('datasets/{}/elements/bundle'.format(self.get_url_prefix()), json_data={'elements': segment})

        self.refresh()

    def close(self, force=False):
        if self.smart_updater is not None:
            self.smart_updater.stop(cancel_pending_jobs=force)

    def sync(self, update_size=True):
        if self.smart_updater is not None:
            while self.smart_updater.queues_busy():
                print("\rTasks pending: {}         ".format(self.smart_updater.tasks_pending), end="", flush=True)
                sleep(1)
            print("\rTasks pending: {}         ".format(self.smart_updater.tasks_pending), end="", flush=True)
        print("\n")

        if update_size:
            self.__update_size()

    def load_from_folder(self, folder):
        self.clear()

        pyfolder = PyFolder(folder)

        if "metadata.json" in pyfolder:
            metadata = self.__load_json(pyfolder)
        elif "metadata.csv" in pyfolder:
            metadata = self.__load_csv(pyfolder)
        else:
            raise FileNotFoundError("No metadata.json or metadata.csv found.")

        content_available = "content" in pyfolder

        if not content_available:
            content_folder = None
            print("Warning: elements do not have content associated.")
        else:
            # We don't want PyFolder to interpret the elements of the dataset.
            content_folder = PyFolder(pyfolder["content"].folder_root, interpret=False)

        elements = []
        batch_size = 0
        for key, values in metadata.items():

            element = {k: v for k, v in values.items() if k != "id"}

            if content_available:
                element['content'] = content_folder[key]
                batch_size += len(element['content'])

            elements.append(element)

            if batch_size >= 1*1024*1024*1024:  # each 1 GB, perform upload.
                print("* Uploading elements...")
                self.add_elements(elements)
                batch_size = 0
                elements = []
                self.sync(update_size=False)

        if len(elements) > 0:
            self.add_elements(elements)

        print("Uploading elements...")
        self.sync()
        print("\rFinished uploading.")

    def __load_json(self, pyfolder):
        return pyfolder["metadata.json"]

    def __load_csv(self, pyfolder):
        content = pyfolder["metadata.csv"].split("\n")
        headers = content[0].split(",")
        data = content[1:]

        result = {}
        for line in data:
            values = {k: v for k, v in zip(headers[1:], line[1:])}
            values['tags'] = values['tags'].split(";")
            result[line[0]] = values

        return result

    def __update_size(self):
        size = self._get_json("datasets/{}/size".format(self.get_url_prefix()))
        self.data['size'] = size



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from dhub.interpreters.interpreter import Interpreter

__author__ = 'Iván de Paz Centeno'


class BinText(Interpreter):

    def cipher(self, text_bytes):
        return text_bytes.decode()

    def decipher(self, text):
        return text.encode()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.
from concurrent.futures import ThreadPoolExecutor
from time import sleep
import requests
from dhub.dhubrc import dhubrc

__author__ = 'Iván de Paz Centeno'


request_pool = ThreadPoolExecutor(2)

class APIWrapper(object):

    def __init__(self, token, api_url=None, token_info=None, server_info=None):
        if api_url is None:
             api_url = dhubrc.get_backend()

        if api_url.endswith("/"): api_url = api_url[:-1]
        self.api_url = api_url
        self.token = token

        if token_info is None:
            self._update_token_info() # server_info is also updated here
        else:
            self.token_info = token_info
            self.server_info = server_info

    def _update_token_info(self):
        try:
            sv_info = request_pool.submit(requests.get, "{}/server".format(self.api_url), params={'_tok': self.token})
            token_info = request_pool.submit(requests.get, "{}/tokens/{}".format(self.api_url, self.token), params={'_tok': self.token})
        except Exception as ex:
            print(ex)
            raise Exception("Backend could not be contacted!")

        self.token_info = token_info.result().json()
        self.server_info = sv_info.result().json()

    def _get_json(self, rel_url, extra_data=None, json_data=None):
        if extra_data is None:
            extra_data = {}

        if json_data is None:
            json_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token

        response = requests.get("{}/{}".format(self.api_url, rel_url), json=json_data, params=data)

        while response.status_code == 429:
            sleep(2)
            response = requests.get("{}/{}".format(self.api_url, rel_url), json=json_data, params=data)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.json()

    def _get_binary(self, rel_url, extra_data=None, json_data=None):
        if extra_data is None:
            extra_data = {}

        if json_data is None:
            json_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token
        response = requests.get("{}/{}".format(self.api_url, rel_url), json=json_data, params=data)

        while response.status_code == 429:
            sleep(2)
            response = requests.get("{}/{}".format(self.api_url, rel_url), json=json_data, params=data)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.content

    def _put_binary(self, rel_url, extra_data=None, binary=None):
        if binary is None:
            binary = b""

        if extra_data is None:
            extra_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token
        response = requests.put("{}/{}".format(self.api_url, rel_url), params=data, data=binary)
        while response.status_code == 429:
            sleep(2)
            response = requests.put("{}/{}".format(self.api_url, rel_url), params=data, data=binary)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.json()

    def _post_json(self, rel_url, extra_data=None, json_data=None):
        if extra_data is None:
            extra_data = {}

        if json_data is None:
            json_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token
        response = requests.post("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)
        while response.status_code == 429:
            sleep(2)
            response = requests.post("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.json()

    def _patch_json(self, rel_url, extra_data=None, json_data=None):
        if extra_data is None:
            extra_data = {}

        if json_data is None:
            json_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token
        response = requests.patch("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)
        while response.status_code == 429:
            sleep(2)
            response = requests.patch("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.json()

    def _delete_json(self, rel_url, extra_data=None, json_data=None):
        if extra_data is None:
            extra_data = {}

        if json_data is None:
            json_data = {}

        if rel_url.startswith("/"):
            rel_url = rel_url[1:]

        data = dict(extra_data)
        data['_tok'] = self.token
        response = requests.delete("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)
        while response.status_code == 429:
            sleep(2)
            response = requests.delete("{}/{}".format(self.api_url, rel_url), params=data, json=json_data)

        if response.status_code not in [200, 201]:
            raise Exception("Failed to communicate with backend: {}".format(response.content.decode()))

        return response.json()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import datetime

__author__ = 'Iván de Paz Centeno'


CACHE_TIME = 60  # seconds for caching elements.

def now():
    return datetime.datetime.now()

def segments(list, size):
    if size <= 0: return []
    list_len = len(list)
    split_count = list_len // size
    extra_split = list_len % size > 0

    for x in range(split_count):
        yield list[x*size: (x+1)*size]

    if extra_split:
        yield list[(split_count)*size:]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# dhub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

import os
import json
from os.path import expanduser

__author__ = 'Iván de Paz Centeno'

class DHubRc(object):
    def __init__(self):
        home = expanduser("~")
        self.token_lookup = {}
        self.backend = "http://localhost:5555/"
        retry = False

        try:
            with open(os.path.join(home, ".dhubrc")) as f:
                rc = json.load(f)

            self.token_lookup = rc['token_lookup']
            self.backend = rc['backend']
        except FileNotFoundError as ex:
            retry = True
            pass

        if not retry:
            return

        try:
            with open("/etc/dhub/config.json") as f:
                rc = json.load(f)

            self.token_lookup = rc['token_lookup']
            self.backend = rc['backend']
        except FileNotFoundError as ex:
            retry = True
            pass

    def lookup_token(self, token_str):
        if token_str is None:
            token_str = "default"

        return self.token_lookup[token_str]

    def get_backend(self):
        return self.backend

dhubrc = DHubRc()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# DHub
# Copyright (C) 2017 Iván de Paz Centeno <ipazc@unileon.es>.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.


import sys
from setuptools import setup, setuptools
from dhub import __version__

__author__ = 'Iván de Paz Centeno'


def readme():
    with open('README.rst') as f:
        return f.read()

if sys.version_info < (3, 4, 1):
    sys.exit('Python < 3.4.1 is not supported!')

setup(name='dhub',
      version=__version__,
      description='ML Data Hub client for managing datasets of MlDataHub from Python3.',
      long_description=readme(),
      url='http://github.com/ipazc/dhub',
      author='Iván de Paz Centeno',
      author_email='ipazc@unileon.es',
      license='GNU GPLv3 or later',
      packages=setuptools.find_packages(),
      install_requires=[
          "pyzip",
          "pyfolder",
          "requests",
          "pillow"
      ],
      classifiers=[
          'Development Status :: 1 - Planning',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Framework :: Flask',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'
      ],
      #entry_points = {
      #    'console_scripts': ['dhub=dhub.entry_point:main'],
      #},
      data_files = [
          ('/etc/dhub', ['dhub/config.json'])
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      keywords="machine-learning mldatahub data hub datasets dataset store client",
      zip_safe=False)
